"""Submodule providing edge prediction models based on Sklearn Models."""
from embiggen.edge_prediction.edge_prediction_sklearn.sklearn_edge_prediction_adapter import SklearnEdgePredictionAdapter
from embiggen.utils.abstract_models import build_init

build_init(
    module_library_names="sklearn",
    formatted_library_name="scikit-learn",
    expected_parent_class=SklearnEdgePredictionAdapter
)


"""Sub-module with utilities to make experiments easier to write."""
from embiggen.utils.abstract_models import (
    AbstractClassifierModel,
    AbstractEmbeddingModel,
    EmbeddingResult,
    AbstractModel,
    get_models_dataframe,
    get_available_models_for_node_label_prediction,
    get_available_models_for_edge_prediction,
    get_available_models_for_edge_label_prediction,
    get_available_models_for_node_embedding,
    abstract_class,
    format_list
)
from embiggen.utils.pipeline import classification_evaluation_pipeline
from embiggen.utils.number_to_ordinal import number_to_ordinal

__all__ = [
    "AbstractClassifierModel",
    "AbstractEmbeddingModel",
    "EmbeddingResult",
    "AbstractModel",
    "classification_evaluation_pipeline",
    "format_list",
    "get_models_dataframe",
    "get_available_models_for_node_label_prediction",
    "get_available_models_for_edge_prediction",
    "get_available_models_for_edge_label_prediction",
    "get_available_models_for_node_embedding",
    "abstract_class",
    "number_to_ordinal",
]



"""Module providing SocioDim implementation."""
from typing import Dict, Any
from ensmallen import Graph
import pandas as pd
from scipy.linalg import eigh
from scipy.sparse.linalg import eigsh
from embiggen.embedders.ensmallen_embedders.ensmallen_embedder import EnsmallenEmbedder
from embiggen.utils import EmbeddingResult


class SocioDimEnsmallen(EnsmallenEmbedder):
    """Class implementing the SocioDim algorithm."""

    def __init__(
        self,
        embedding_size: int = 100,
        use_sparse_reduce: bool = True,
        enable_cache: bool = False
    ):
        """Create new SocioDim method.

        Parameters
        --------------------------
        embedding_size: int = 100
            Dimension of the embedding.
        use_sparse_reduce: bool = True
            Whether to use the sparse reduce or the dense reduce for
            the computation of eigenvectors.
            For both reduce mechanisms, we are using LAPACK implementations.
            For some currently unknown reason, their implementation using
            a sparse reduce, even on dense matrices, yields better results.
        enable_cache: bool = False
            Whether to enable the cache, that is to
            store the computed embedding.
        """
        self._use_sparse_reduce = use_sparse_reduce
        super().__init__(
            embedding_size=embedding_size,
            enable_cache=enable_cache,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters of the model."""
        return {
            **super().parameters(),
            **dict(
                use_sparse_reduce=self._use_sparse_reduce,
            )
        }

    def _fit_transform(
        self,
        graph: Graph,
        return_dataframe: bool = True,
    ) -> EmbeddingResult:
        """Return node embedding."""
        if self._use_sparse_reduce:
            embedding = eigsh(
                graph.get_dense_modularity_matrix(),
                k=self._embedding_size,
                which="LM",
                return_eigenvectors=True
            )[1]
        else:
            number_of_nodes = graph.get_number_of_nodes()
            embedding = eigh(
                graph.get_dense_modularity_matrix(),
                eigvals=(
                    number_of_nodes-self._embedding_size,
                    number_of_nodes-1
                )
            )[1]

        if return_dataframe:
            node_names = graph.get_node_names()
            embedding = pd.DataFrame(
                embedding,
                index=node_names
            )
        return EmbeddingResult(
            embedding_method_name=self.model_name(),
            node_embeddings=embedding
        )

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model."""
        return "SocioDim"

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return False

    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        return False



"""Module providing abstract Node2Vec implementation."""
from typing import Dict, Any
from ensmallen import Graph
import pandas as pd
from ensmallen import models
from embiggen.embedders.ensmallen_embedders.ensmallen_embedder import EnsmallenEmbedder
from embiggen.utils import EmbeddingResult


class WeightedSPINE(EnsmallenEmbedder):
    """Abstract class for Node2Vec algorithms."""

    def __init__(
        self,
        embedding_size: int = 100,
        use_edge_weights_as_probabilities: bool = False,
        verbose: bool = False,
        enable_cache: bool = False
    ):
        """Create new abstract Node2Vec method.

        Parameters
        --------------------------
        embedding_size: int = 100
            Dimension of the embedding.
        use_edge_weights_as_probabilities: bool = False
            Whether to treat the weights as probabilities.
        verbose: bool = False
            Whether to show loading bars.
        enable_cache: bool = False
            Whether to enable the cache, that is to
            store the computed embedding.
        """
        self._verbose = verbose
        self._model = models.WeightedSPINE(
            embedding_size=embedding_size,
            use_edge_weights_as_probabilities=use_edge_weights_as_probabilities,
        )

        super().__init__(
            embedding_size=embedding_size,
            enable_cache=enable_cache
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            embedding_size=5,
        )

    def _fit_transform(
        self,
        graph: Graph,
        return_dataframe: bool = True,
    ) -> EmbeddingResult:
        """Return node embedding."""
        node_embedding = self._model.fit_transform(
            graph,
            verbose=self._verbose,
        ).T
        if return_dataframe:
            node_embedding = pd.DataFrame(
                node_embedding,
                index=graph.get_node_names()
            )
        return EmbeddingResult(
            embedding_method_name=self.model_name(),
            node_embeddings=node_embedding
        )

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model."""
        return "WeightedSPINE"

    @classmethod
    def requires_edge_weights(cls) -> bool:
        return True

    @classmethod
    def requires_positive_edge_weights(cls) -> bool:
        return True

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return False

    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        return False



from typing import List, Optional, Union, Dict, Tuple
import pandas as pd
import numpy as np
from ensmallen import models, Graph


class DAGResnik:

    def __init__(self, verbose: bool = True):
        """Create new Resnik similarity model."""
        self._model = models.DAGResnik(verbose)

    def fit(
        self,
        graph: Graph,
        node_counts: Dict[str, float],
        node_frequencies: Optional[np.ndarray] = None,
    ):
        """Fit the Resnik similarity model.

        Parameters
        --------------------
        graph: Graph
            The graph to run similarities on.
        node_counts: Dict[str, float]
            Counts to compute the terms frequencies.
        node_frequencies: Optional[np.ndarray] = None
            Optional vector of node frequencies.
        """
        self._model.fit(
            graph,
            node_counts=node_counts,
            node_frequencies=node_frequencies
        )

    def get_similarity_from_node_id(
        self,
        first_node_id: int,
        second_node_id: int
    ) -> float:
        """Return the similarity between the two provided nodes.

        Arguments
        --------------------
        first_node_id: int
            The first node for which to compute the similarity.
        second_node_id: int
            The second node for which to compute the similarity.
        """
        return self._model.get_similarity_from_node_id(first_node_id, second_node_id)

    def get_similarity_from_node_ids(
        self,
        first_node_ids: List[int],
        second_node_ids: List[int],
        minimum_similarity: Optional[float] = 0.0,
    ) -> np.ndarray:
        """Return the similarity between the two provided nodes.

        Arguments
        --------------------
        first_node_ids: List[int]
            The first node for which to compute the similarity.
        second_node_ids: List[int]
            The second node for which to compute the similarity.
        """
        return self._model.get_similarity_from_node_ids(first_node_ids, second_node_ids, minimum_similarity)

    def get_similarity_from_node_name(
        self,
        first_node_name: str,
        second_node_name: str
    ) -> float:
        """Return the similarity between the two provided nodes.

        Arguments
        --------------------
        first_node_name: str
            The first node for which to compute the similarity.
        second_node_name: str
            The second node for which to compute the similarity.
        """
        return self._model.get_similarity_from_node_name(first_node_name, second_node_name)

    def get_similarity_from_node_names(
        self,
        first_node_names: List[str],
        second_node_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
    ) -> np.ndarray:
        """Return the similarity between the two provided nodes.

        Arguments
        --------------------
        first_node_names: List[str]
            The first node for which to compute the similarity.
        second_node_names: List[str]
            The second node for which to compute the similarity.
        """
        return self._model.get_similarity_from_node_names(first_node_names, second_node_names, minimum_similarity)

    def get_pairwise_similarities(
        self,
        graph: Optional[Graph] = None,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities on the provided graph.

        Parameters
        --------------------
        graph: Optional[Graph] = None
            The graph to run similarities on.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        similarities = self._model.get_pairwise_similarities()

        if return_similarities_dataframe and graph is not None:
            similarities = pd.DataFrame(
                similarities,
                columns=graph.get_node_names(),
                index=graph.get_node_names(),
            )

        return similarities

    def get_similarities_from_graph(
        self,
        graph: Graph,
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run similarities on.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        (edge_node_ids, similarities) = self._model.get_similarities_from_graph(
            graph,
            minimum_similarity
        )

        if return_similarities_dataframe:
            if not graph.has_edges():
                similarities = pd.DataFrame({
                    "similarity": [],
                    "source": [],
                    "destination": [],
                })
            else:
                similarities = pd.DataFrame(
                    {
                        "similarity": similarities,
                        "source": edge_node_ids[:, 0],
                        "destination": edge_node_ids[:, 1],
                    },
                )

        return similarities

    def get_similarities_from_bipartite_graph_from_edge_node_ids(
        self,
        graph: Graph,
        source_node_ids: List[int],
        destination_node_ids: List[int],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_ids: List[int]
            The source nodes of the bipartite graph.
        destination_node_ids: List[int]
            The destination nodes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_bipartite_graph_from_edge_node_ids(
                source_node_ids=source_node_ids,
                destination_node_ids=destination_node_ids,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )

    def get_similarities_from_bipartite_graph_from_edge_node_names(
        self,
        graph: Graph,
        source_node_names: List[str],
        destination_node_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_names: List[str]
            The source nodes of the bipartite graph.
        destination_node_names: List[str]
            The destination nodes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_bipartite_graph_from_edge_node_names(
                source_node_names=source_node_names,
                destination_node_names=destination_node_names,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )

    def get_similarities_from_bipartite_graph_from_edge_node_prefixes(
        self,
        source_node_prefixes: List[str],
        destination_node_prefixes: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, Tuple[List[Tuple[str, str]], np.ndarray]]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        source_node_prefixes: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_prefixes: List[str]
            The destination node prefixes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        edge_node_names, similarities = self._model.get_similarity_from_node_prefixes(
            first_node_prefixes=source_node_prefixes,
            second_node_prefixes=destination_node_prefixes,
            minimum_similarity=minimum_similarity,
        )

        if not return_similarities_dataframe:
            return edge_node_names, similarities
        
        edge_node_names = pd.DataFrame(
            edge_node_names,
            columns=["source", "destination"],
        )

        edge_node_names["similarity"] = similarities
        
        return edge_node_names

    def get_similarities_from_bipartite_graph_from_edge_node_types(
        self,
        graph: Graph,
        source_node_types: List[str],
        destination_node_types: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_types: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_types: List[str]
            The destination node prefixes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_bipartite_graph_from_edge_node_types(
                source_node_types=source_node_types,
                destination_node_types=destination_node_types,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )

    def get_similarities_from_clique_graph_from_node_ids(
        self,
        graph: Graph,
        node_ids: List[int],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_ids: List[int]
            The nodes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_clique_graph_from_node_ids(
                node_ids=node_ids,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )

    def get_similarities_from_clique_graph_from_node_names(
        self,
        graph: Graph,
        node_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_names: List[str]
            The nodes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_clique_graph_from_node_names(
                node_names=node_names,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )

    def get_similarities_from_clique_graph_from_node_prefixes(
        self,
        node_prefixes: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, Tuple[List[Tuple[str, str]], np.ndarray]]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        node_prefixes: List[str]
            The node prefixes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_bipartite_graph_from_edge_node_prefixes(
            source_node_prefixes=node_prefixes,
            destination_node_prefixes=node_prefixes,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )        


    def get_similarities_from_clique_graph_from_node_types(
        self,
        graph: Graph,
        node_types: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_types: List[str]
            The node prefixes of the bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as it weights much less.
        """
        return self.get_similarities_from_graph(
            graph.build_clique_graph_from_node_types(
                node_types=node_types,
                directed=True
            ),
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe
        )
